/**
 * @file
 *
 * Operators and utility classes for different database execution models.
 * BaseOperator.h provides documentation of execution model interfaces.
 *
 * @author Henning Funke <henning.funke@cs.tu-dortmund.de>
 *
 * (c) 2019 TU Dortmund, DBIS Group
 */

#pragma once

#include <iostream>
#include <pthread.h>
#include <thread>
#include <cstdlib>

#include "BaseOperator.h"
#include "DBData.h"


/* forward declarations */
namespace llvm {
    class Value;
}

class JitContext;

class CJitContext;

/**
 * @brief Methods to drive the plan execution for pull-based execution models.
 */
class PullDriver {
public:
    /**
     * @brief Execute query plan with Volcano (Tuple-at-a-time) and write result.
     */
    static void volcano ( RelOperator* node, Relation* result ) {
        node->open();
        Tuple* t = node->next();
        size_t outLen=0;
        Tuple* r = result->r;
        while ( t != nullptr) {
            r[outLen++] = *t;
            t = node->next();
        }
        result->len = outLen;
        node->close();
    }

    /**
     * @brief Execute query plan with Vector-at-a-time and write result.
     */
    static void vector ( RelOperator* node, Relation* result ) {
        node->openVec();
        Relation* resVec = node->nextVec();
        size_t outLen=0;
        Tuple* r;
        Tuple* rOut = result->r;
        while ( resVec != nullptr ) {
            r = resVec->r;
            for ( size_t i=0; i < resVec->len; i++ ) {
                rOut[outLen++] = r[i];
            } 
            resVec = node->nextVec();
        }
        result->len = outLen;
        node->closeVec();
    }
};




/**
 * @brief Operator for scanning a relation.
 */
class ScanOp : public RelOperator {
protected:
    Tuple* table;
    size_t tableSize;
    /* volcano (and vector-at-a-time) */
    size_t cursor;
    /* operator-at-a-time */
    Relation oCol;
    /* vector-at-a-time */
    Relation oVec;
    
public:
    ScanOp ( Tuple *tab, size_t n ) : RelOperator ( nullptr ) {
        this->table = tab;
        this->tableSize = n;
        this->oCol = allocateRelation ( n );
    }

    virtual ~ScanOp() {
        freeRelation ( this->oCol );
    }
    
    virtual size_t getSize () {
        return tableSize;
    }
    
    virtual void open();
    virtual Tuple* next();
    virtual void close();

    virtual Relation getRelation();

    virtual void openVec();
    virtual Relation* nextVec();
    virtual void closeVec();
};


/**
 * @brief Operator for selection.
 */
class SelectionOp : public RelOperator {

public:
    enum PredicateType { SMALLER, EQUALS, EQUALS_NOT };

protected:
    SelectionOp::PredicateType type;
    int compareConstant;
    /* vector-at-a-time */
    Relation oVec;

public:
    SelectionOp( PredicateType type, int compareConstant, RelOperator* child ) : RelOperator ( child ) {
        this->type = type;
        this->compareConstant = compareConstant;
    }

    virtual ~SelectionOp() {};
    
    virtual size_t getSize () {
        return child->getSize();
    }
 
    virtual void open();
    virtual Tuple* next();
    virtual void close();

    virtual Relation getRelation();

    virtual void openVec();
    virtual Relation* nextVec();
    virtual void closeVec();
};


/**
 * @brief Aggregation operator.
 */
class AggregationOp : public RelOperator {
public:
    enum ReduceType { COUNT, SUM };

protected:
    AggregationOp::ReduceType type;
    Tuple countTuple;
    bool hasMoreTuples;

    /* operator-at-a-time */
    Relation oCol;
    /* vector-at-a-time */
    Relation oVec;
    /* Jit compilation */
    llvm::Value *sumStackPtr;

public:
    AggregationOp ( ReduceType type, RelOperator* child ) : RelOperator ( child ) {
        this->oCol = allocateRelation ( 1 );
        this->type = type;
    }
    
    virtual ~AggregationOp() {
        freeRelation ( this->oCol );
    }
    
    virtual size_t getSize () {
        return 1;
    }

    virtual void open();
    virtual Tuple* next();
    virtual void close();

    virtual Relation getRelation();

    virtual void openVec();
    virtual Relation* nextVec();
    virtual void closeVec();
    
    /**
     * @brief Called by produceC after aggregate is finished.
     */
    //void consumeAggregateC ( CJitContext& ctx );
    
    /**
     * @brief Called by produce after aggregate is finished.
     */
    //void consumeAggregate ( JitContext& ctx );
};

/**
 * @brief Materialization operator for JIT-compilation.
 * This operator is responsible for writing the current tuple in JIT-code 
 * to the result relation. It is attached to plans as root to trigger materialization
 * of register-contents to memory. Other execution models materialize to memory without the operator.
 */
class MaterializeOp : public RelOperator {
public:
    MaterializeOp ( RelOperator* child ) : RelOperator ( child ) {}
    
    virtual size_t getSize () { return 0; };

/*    
    virtual void produceC ( CJitContext& ctx );
    virtual void consumeC ( CJitContext& ctx );
    void consumeReturnC ( CJitContext& ctx );

    virtual void produce ( JitContext& ctx );
    virtual void consume ( JitContext& ctx );
    void consumeReturn ( JitContext& ctx );
*/  
  
    /* unused */
    virtual void open() {};
    virtual Tuple* next() { return nullptr; };
    virtual void close() {};
    virtual Relation getRelation() { return Relation(); };
    virtual void openVec() {};
    virtual Relation* nextVec() { return nullptr; };
    virtual void closeVec() {};

};
